@@ -434,6 +434,7 @@ ResponseList Controller::ComputeResponseList(std::atomic_bool& shut_down,
     // order consistently across workers.
     for (auto& response : response_list.responses()) {
       if ((response.response_type() == Response::ResponseType::ALLREDUCE ||
+           response.response_type() == Response::ResponseType::ALLGATHER ||
            response.response_type() == Response::ResponseType::ADASUM ||
            response.response_type() == Response::ResponseType::ALLTOALL) &&
           (int)response.devices().size() == size_) {
@@ -985,7 +986,7 @@ void Controller::SynchronizeTimelineEnabled() {
   timeline_enabled_ = timeline_enabled_pending_;
 }
 
-bool Controller::TimeLineEnabled() {
+bool Controller::TimelineEnabled() {
   std::lock_guard<std::recursive_mutex> guard(timeline_mutex_);
   return timeline_enabled_;
 }
