@@ -23,7 +23,17 @@
 #include "gloo/rendezvous/context.h"
 #include "gloo/rendezvous/file_store.h"
 #include "gloo/rendezvous/prefix_store.h"
+
+#ifdef __linux__
 #include "gloo/transport/tcp/device.h"
+using attr = gloo::transport::tcp::attr;
+constexpr auto CreateDevice = gloo::transport::tcp::CreateDevice;
+#else
+// Use uv on macOS as TCP requires epoll (Linux-only)
+#include "gloo/transport/uv/device.h"
+using attr = gloo::transport::uv::attr;
+constexpr auto CreateDevice = gloo::transport::uv::CreateDevice;
+#endif
 
 #if HAVE_MPI
 #include "gloo/mpi/context.h"
@@ -36,23 +46,6 @@
 namespace horovod {
 namespace common {
 
-// Horovod Gloo rendezvous knobs.
-#define HOROVOD_GLOO_TIMEOUT_SECONDS "HOROVOD_GLOO_TIMEOUT_SECONDS"
-#define HOROVOD_GLOO_RENDEZVOUS_ADDR "HOROVOD_GLOO_RENDEZVOUS_ADDR"
-#define HOROVOD_GLOO_RENDEZVOUS_PORT "HOROVOD_GLOO_RENDEZVOUS_PORT"
-#define HOROVOD_GLOO_GLOBAL_PREFIX "global"
-#define HOROVOD_GLOO_LOCAL_PREFIX "local_"
-#define HOROVOD_GLOO_CROSS_PREFIX "cross_"
-#define HOROVOD_GLOO_GET_RANK_AND_SIZE "rank_and_size"
-#define HOROVOD_HOSTNAME "HOROVOD_HOSTNAME"
-#define HOROVOD_RANK "HOROVOD_RANK"
-#define HOROVOD_SIZE "HOROVOD_SIZE"
-#define HOROVOD_LOCAL_RANK "HOROVOD_LOCAL_RANK"
-#define HOROVOD_LOCAL_SIZE "HOROVOD_LOCAL_SIZE"
-#define HOROVOD_CROSS_RANK "HOROVOD_CROSS_RANK"
-#define HOROVOD_CROSS_SIZE "HOROVOD_CROSS_SIZE"
-#define HOROVOD_ELASTIC "HOROVOD_ELASTIC"
-
 int ParseNextInt(std::stringstream& ss) {
   assert(ss.good());
 
@@ -80,7 +73,8 @@ std::shared_ptr<gloo::Context> Rendezvous(const std::string& prefix,
     store.reset(new MemoryStore());
   }
   LOG(DEBUG) << prefix << " rendezvous started for rank=" << rank << ", size=" << size
-             << ", dev={" << dev->str() << "}";
+             << ", dev={" << dev->str() << "}, timeout="
+             << std::to_string(std::chrono::duration_cast<std::chrono::seconds>(timeout).count());
 
   auto context = std::make_shared<gloo::rendezvous::Context>(rank, size);
   context->setTimeout(timeout);
@@ -98,10 +92,10 @@ void GlooContext::InitializeFromMPI(MPIContext& mpi_ctx,
 
   // TODO(sihan): Add support for multiple interfaces:
   //  https://github.com/facebookincubator/gloo/issues/190
-  gloo::transport::tcp::attr attr;
-  attr.iface = gloo_iface;
-  attr.ai_family = AF_UNSPEC;
-  auto dev = gloo::transport::tcp::CreateDevice(attr);
+  attr device_attr;
+  device_attr.iface = gloo_iface;
+  device_attr.ai_family = AF_UNSPEC;
+  auto dev = CreateDevice(device_attr);
   auto timeout = GetTimeoutFromEnv();
 
   auto context =
@@ -129,14 +123,14 @@ void GlooContext::Initialize(const std::string& gloo_iface) {
     return;
   }
 
-  // Create a tcp device for communication
+  // Create a device for communication
   // TODO(sihan): Add support for multiple interfaces:
   //  https://github.com/facebookincubator/gloo/issues/190
-  gloo::transport::tcp::attr attr;
-  attr.iface = gloo_iface;
+  attr device_attr;
+  device_attr.iface = gloo_iface;
 
-  attr.ai_family = AF_UNSPEC;
-  auto dev = gloo::transport::tcp::CreateDevice(attr);
+  device_attr.ai_family = AF_UNSPEC;
+  auto dev = CreateDevice(device_attr);
   auto timeout = GetTimeoutFromEnv();
 
   auto host_env = std::getenv(HOROVOD_HOSTNAME);
