@@ -40,6 +40,7 @@
 #include "parameter_manager.h"
 #include "timeline.h"
 #include "utils/env_parser.h"
+#include "nvtx_op_range.h"
 
 #if HAVE_MPI
 #define OMPI_SKIP_MPICXX
@@ -260,7 +261,7 @@ void PerformOperation(Response response, HorovodGlobalState& state) {
                                                              state.joined);
 
     for (auto& e : entries) {
-      timeline.Start(e.tensor_name, response.response_type());
+      timeline.Start(e.tensor_name, response.response_type(), e.tensor->size());
     }
 
     if (entries.size() > 1) {
@@ -278,10 +279,7 @@ void PerformOperation(Response response, HorovodGlobalState& state) {
         LOG(DEBUG, horovod_global.controller->GetRank()) << "InitializeBuffer Failed";
         for (auto& e : entries) {
           timeline.End(e.tensor_name, nullptr);
-          // Callback can be null if the rank sent Join request.
-          if (e.callback != nullptr) {
-            e.callback(status);
-          }
+          e.FinishWithCallback(status);
         }
         return;
       }
@@ -325,10 +323,7 @@ void PerformOperation(Response response, HorovodGlobalState& state) {
   if (!status.in_progress()) {
     for (auto& e : entries) {
       timeline.End(e.tensor_name, status.ok() ? e.output : nullptr);
-      // Callback can be null if the rank sent Join request.
-      if (e.callback != nullptr) {
-        e.callback(status);
-      }
+      e.FinishWithCallback(status);
     }
   }
 }
@@ -417,20 +412,29 @@ void BackgroundThreadLoop(HorovodGlobalState& state) {
   gpu_context.finalizer_thread_pool.create(state.num_nccl_streams);
 #endif
 
+#if HAVE_NVTX
+  if (GetBoolEnvOrDefault(HOROVOD_DISABLE_NVTX_RANGES, false)) {
+    NvtxOpRange::nvtx_ops_handle.Disable();
+    horovod_global.timeline.DisableNvtx();
+  }
+#endif // HAVE_NVTX
+
   // Open the timeline file on coordinator.
-  auto timeline_env = std::getenv(HOROVOD_TIMELINE);
-  auto horovod_timeline = timeline_env != nullptr ? std::string(timeline_env) : std::string("");
   bool should_enable_timeline = false;
-  if (is_coordinator) {
-    state.timeline.Initialize(horovod_timeline,
-                              static_cast<unsigned int>(size));
+  auto timeline_env = std::getenv(HOROVOD_TIMELINE);
+  if (timeline_env) {
+    if (is_coordinator) {
+      auto horovod_timeline = std::string(timeline_env);
+      if (horovod_timeline != "DYNAMIC") {
+        state.timeline.Initialize(horovod_timeline,
+                                  static_cast<unsigned int>(size));
+      } else {
+        state.timeline.Initialize("",
+                                  static_cast<unsigned int>(size));
+      }
+    }
+    should_enable_timeline = true;
   }
-
-  should_enable_timeline = true;
-  if (horovod_timeline == "DISABLED") {
-    should_enable_timeline = false;
-  }
-
   state.controller->SetTimelineEnabled(should_enable_timeline);
 
   ParseStallInspectorFromEnv(state.controller->GetStallInspector());
@@ -564,11 +568,7 @@ void BackgroundThreadLoop(HorovodGlobalState& state) {
 
   // Notify all outstanding operations that Horovod has been shut down
   // and finalize tensor queue.
-  std::vector<StatusCallback> callbacks;
-  horovod_global.tensor_queue.FinalizeTensorQueue(callbacks);
-  for (auto& cb : callbacks) {
-    cb(SHUT_DOWN_ERROR);
-  }
+  horovod_global.tensor_queue.FinalizeTensorQueue(SHUT_DOWN_ERROR);
 
 #if HAVE_GPU
   gpu_context.Finalize();
@@ -737,9 +737,12 @@ bool horovod_is_initialized() {
   return horovod_global.initialization_done;
 }
 
-bool horovod_start_timeline(const char* file_name, bool mark_cycles) {
+int horovod_start_timeline(const char* file_name, bool mark_cycles) {
   if (!horovod_global.initialization_done) {
-    return false;
+    return -1;
+  }
+  if (!horovod_global.controller->TimelineEnabled()) {
+    return -2;
   }
   bool is_coordinator = horovod_global.controller->IsCoordinator();
   if (is_coordinator) {
@@ -747,22 +750,22 @@ bool horovod_start_timeline(const char* file_name, bool mark_cycles) {
     horovod_global.timeline.SetPendingTimelineFile(std::string(file_name));
   }
   horovod_global.controller->SetMarkCyclesInTimelinePending(mark_cycles);
-  return true;
+  return 1;
 }
 
-bool horovod_stop_timeline() {
+int horovod_stop_timeline() {
   if (!horovod_global.initialization_done) {
-    return false;
+    return -1;
   }
   if(!horovod_global.controller->TimelineEnabledPending()){
     LOG(INFO) << " Timeline is already stopped. Please start timeline before stopping it.";
-    return true;
+    return 1;
   }
   bool is_coordinator = horovod_global.controller->IsCoordinator();
   if (is_coordinator) {
       horovod_global.timeline.SetPendingTimelineFile(std::string(""));
   }
-  return true;
+  return 1;
 }
 
 int horovod_rank() {
@@ -779,6 +782,13 @@ int horovod_local_rank() {
   return horovod_global.controller->GetLocalRank();
 }
 
+int horovod_cross_rank() {
+  if (!horovod_global.initialization_done) {
+    return -1;
+  }
+  return horovod_global.controller->GetCrossRank();
+}
+
 int horovod_size() {
   if (!horovod_global.initialization_done) {
     return -1;
@@ -793,6 +803,13 @@ int horovod_local_size() {
   return horovod_global.controller->GetLocalSize();
 }
 
+int horovod_cross_size() {
+  if (!horovod_global.initialization_done) {
+    return -1;
+  }
+  return horovod_global.controller->GetCrossSize();
+}
+
 bool horovod_is_homogeneous() {
   return horovod_global.controller->IsHomogeneous();
 }
@@ -996,14 +1013,30 @@ Status EnqueueTensorAllreduces(std::vector<std::shared_ptr<OpContext>>& contexts
     e.callback = std::move(callbacks[n]);
 
     entries.push_back(std::move(e));
+  }
 
+  // Start appropriate NVTX range
+  if (tensors.size() == 1) {
+    auto& e = entries[0];
+    e.nvtx_op_range.Start(RegisteredNvtxOp::HorovodAllreduce, e.tensor->size());
+  } else {
+    auto total_size =
+        std::accumulate(entries.begin(), entries.end(), 0ll,
+                        [](int64_t size_sum, const TensorTableEntry& e) {
+                          return size_sum + e.tensor->size();
+                        });
+    SharedNvtxOpRange range;
+    range.Start(RegisteredNvtxOp::HorovodGroupedAllreduce, total_size);
+    for (auto& e : entries) {
+      e.nvtx_op_range = range;
+    }
   }
 
   std::string tensors_enqueued;
   for (const auto& n : names) {
     tensors_enqueued += n + "; ";
   }
-  LOG(TRACE) << "Enqueing " << tensors_enqueued;
+  LOG(TRACE, horovod_global.controller->GetRank()) << "Enqueued " << tensors_enqueued;
 
   // Only create groups larger than 1 tensor, unless disable_group_fusion is requested.
   // In that case, even single tensor groups are created to enforce disabling fusion.
@@ -1027,7 +1060,7 @@ Status EnqueueTensorAllreduces(std::vector<std::shared_ptr<OpContext>>& contexts
 Status EnqueueTensorAllgather(std::shared_ptr<OpContext> context,
                               std::shared_ptr<Tensor> tensor,
                               std::shared_ptr<ReadyEvent> ready_event,
-                              const std::string name, const int device,
+                              const std::string& name, const int device,
                               StatusCallback callback) {
   Request message;
   message.set_request_rank(horovod_global.controller->GetRank());
@@ -1046,6 +1079,7 @@ Status EnqueueTensorAllgather(std::shared_ptr<OpContext> context,
   e.ready_event = ready_event;
   e.device = device;
   e.callback = callback;
+  e.nvtx_op_range.Start(RegisteredNvtxOp::HorovodAllgather, e.tensor->size());
 
   if (horovod_global.shut_down) {
     return SHUT_DOWN_ERROR;
@@ -1063,7 +1097,7 @@ Status EnqueueTensorBroadcast(std::shared_ptr<OpContext> context,
                               std::shared_ptr<Tensor> tensor,
                               std::shared_ptr<Tensor> output, int root_rank,
                               std::shared_ptr<ReadyEvent> ready_event,
-                              const std::string name, const int device,
+                              const std::string& name, const int device,
                               StatusCallback callback) {
   Request message;
   message.set_request_rank(horovod_global.controller->GetRank());
@@ -1085,6 +1119,7 @@ Status EnqueueTensorBroadcast(std::shared_ptr<OpContext> context,
   e.ready_event = ready_event;
   e.device = device;
   e.callback = callback;
+  e.nvtx_op_range.Start(RegisteredNvtxOp::HorovodBroadcast, e.tensor->size());
 
   if (horovod_global.shut_down) {
     return SHUT_DOWN_ERROR;
@@ -1102,7 +1137,7 @@ Status EnqueueTensorAlltoall(std::shared_ptr<OpContext> context,
                              std::shared_ptr<Tensor> tensor,
                              std::shared_ptr<Tensor> splits,
                              std::shared_ptr<ReadyEvent> ready_event,
-                             const std::string name, const int device,
+                             const std::string& name, const int device,
                              StatusCallback callback) {
   // Check arguments
   if (splits->shape().dims() > 1) {
@@ -1129,6 +1164,7 @@ Status EnqueueTensorAlltoall(std::shared_ptr<OpContext> context,
   e.ready_event = ready_event;
   e.device = device;
   e.callback = callback;
+  e.nvtx_op_range.Start(RegisteredNvtxOp::HorovodAlltoall, e.tensor->size());
 
   int64_t splits_first_dim = splits->shape().dim_size(0);
   int64_t tensor_first_dim = tensor->shape().dim_size(0);
@@ -1165,7 +1201,7 @@ Status EnqueueTensorAlltoall(std::shared_ptr<OpContext> context,
 // must be running before this function is called.
 Status EnqueueJoin(std::shared_ptr<OpContext> context,
                    std::shared_ptr<ReadyEvent> ready_event,
-                   const std::string name, const int device,
+                   const std::string& name, const int device,
                    StatusCallback callback) {
   Request message;
   message.set_request_rank(horovod_global.controller->GetRank());
