@@ -115,7 +115,7 @@ Status EnqueueTensorAllreduce(std::shared_ptr<OpContext> context,
                               std::shared_ptr<Tensor> tensor,
                               std::shared_ptr<Tensor> output,
                               std::shared_ptr<ReadyEvent> ready_event,
-                              std::string name, const int device,
+                              std::string name, int device,
                               StatusCallback callback,
                               ReduceOp reduce_op = ReduceOp::SUM,
                               double prescale_factor = 1.0,
@@ -126,7 +126,7 @@ Status EnqueueTensorAllreduces(std::vector<std::shared_ptr<OpContext>>& contexts
                                std::vector<std::shared_ptr<Tensor>>& outputs,
                                std::vector<std::shared_ptr<ReadyEvent>>& ready_events,
                                std::vector<std::string>& names,
-                               const int device,
+                               int device,
                                std::vector<StatusCallback>& callbacks,
                                ReduceOp reduce_op = ReduceOp::SUM,
                                double prescale_factor = 1.0,
@@ -135,27 +135,27 @@ Status EnqueueTensorAllreduces(std::vector<std::shared_ptr<OpContext>>& contexts
 Status EnqueueTensorAllgather(std::shared_ptr<OpContext> context,
                               std::shared_ptr<Tensor> tensor,
                               std::shared_ptr<ReadyEvent> ready_event,
-                              const std::string name, const int device,
+                              const std::string& name, int device,
                               StatusCallback callback);
 
 Status EnqueueTensorBroadcast(std::shared_ptr<OpContext> context,
                               std::shared_ptr<Tensor> tensor,
                               std::shared_ptr<Tensor> output, int root_rank,
                               std::shared_ptr<ReadyEvent> ready_event,
-                              const std::string name, const int device,
+                              const std::string& name, int device,
                               StatusCallback callback);
 
 Status EnqueueTensorAlltoall(std::shared_ptr<OpContext> context,
                              std::shared_ptr<Tensor> tensor,
                              std::shared_ptr<Tensor> splits,
                              std::shared_ptr<ReadyEvent> ready_event,
-                             const std::string name, const int device,
+                             const std::string& name, int device,
                              StatusCallback callback);
 
 Status EnqueueJoin(std::shared_ptr<OpContext> context,
-                              std::shared_ptr<ReadyEvent> ready_event,
-                              const std::string name, const int device,
-                              StatusCallback callback);
+                   std::shared_ptr<ReadyEvent> ready_event,
+                   const std::string& name, int device,
+                   StatusCallback callback);
 
 } // namespace common
 } // namespace horovod
