@@ -25,6 +25,25 @@
 namespace horovod {
 namespace common {
 
+static Response::ResponseType RequestTypeToResponseType(Request::RequestType value) {
+  switch (value) {
+    case Request::RequestType::ALLREDUCE:
+      return Response::ResponseType::ALLREDUCE;
+    case Request::RequestType::ALLGATHER:
+      return Response::ResponseType::ALLGATHER;
+    case Request::RequestType::BROADCAST:
+      return Response::ResponseType::BROADCAST;
+    case Request::RequestType::JOIN:
+      return Response::ResponseType::JOIN;
+    case Request::RequestType::ADASUM:
+      return Response::ResponseType::ADASUM;
+    case Request::RequestType::ALLTOALL:
+      return Response::ResponseType::ALLTOALL;
+    default:
+      throw std::logic_error("No corresponding ResponseType for provided RequestType.");
+  }
+}
+
 void ResponseCache::clear() {
   bits_outdated_ = false;
   cache_.clear();
@@ -59,7 +78,8 @@ ResponseCache::CacheState ResponseCache::cached(const Request& message) const {
             cache_params.dtype == message.tensor_type() &&
             cache_params.shape == message.tensor_shape() &&
             cache_response.prescale_factor() == message.prescale_factor() &&
-            cache_response.postscale_factor() == message.postscale_factor())
+            cache_response.postscale_factor() == message.postscale_factor() &&
+            cache_response.response_type() == RequestTypeToResponseType(message.request_type()))
                ? CacheState::HIT
                : CacheState::INVALID;
   } else {
@@ -94,7 +114,8 @@ ResponseCache::cached(const Response& response,
     return (cache_params.device == params.device &&
             cache_params.dtype == params.dtype && same_shape &&
             cache_response.prescale_factor() == response.prescale_factor() &&
-            cache_response.postscale_factor() == response.postscale_factor())
+            cache_response.postscale_factor() == response.postscale_factor() &&
+            cache_response.response_type() == response.response_type())
                ? CacheState::HIT
                : CacheState::INVALID;
   } else {
@@ -171,6 +192,7 @@ void ResponseCache::put(const Response& response, TensorQueue& tensor_queue, boo
                                               joined);
   }
 
+  int64_t global_size = response.tensor_sizes().size() / response.tensor_names().size();
   // If response is fused, split back into individual responses
   if (response.tensor_names().size() > 1) {
     int64_t i = 0;
@@ -179,7 +201,11 @@ void ResponseCache::put(const Response& response, TensorQueue& tensor_queue, boo
       new_response.add_tensor_name(name);
       new_response.set_response_type(response.response_type());
       new_response.set_devices(response.devices());
-      new_response.add_tensor_size(response.tensor_sizes()[i]);
+      // For allreduce, adasum and alltoall, tensor_sizes are the num_elements of response
+      // For allgather, tensor_sizes are the first dim of all ranks
+      for (int64_t j = 0; j < global_size; ++j) {
+        new_response.add_tensor_size(response.tensor_sizes()[i * global_size + j]);
+      }
       new_response.set_tensor_type(response.tensor_type());
       new_response.set_prescale_factor(response.prescale_factor());
       new_response.set_postscale_factor(response.postscale_factor());
