@@ -19,6 +19,10 @@
 #include <chrono>
 #include <sstream>
 
+#if HAVE_NVTX
+#include <nvtx3/nvToolsExt.h>
+#endif
+
 #include "logging.h"
 
 namespace horovod {
@@ -31,7 +35,7 @@ TimelineWriter::TimelineWriter() {
   pending_status_ = false;
 }
 
-void TimelineWriter::SetPendingTimelineFile(std::string filename) {
+void TimelineWriter::SetPendingTimelineFile(const std::string& filename) {
   {
     std::lock_guard<std::recursive_mutex> guard(writer_mutex_);
     if (cur_filename_ == filename) {
@@ -50,7 +54,7 @@ void TimelineWriter::SetPendingTimelineFile(std::string filename) {
         return;
       }
     }
-    if (filename == "") {
+    if (filename.empty()) {
       LOG(DEBUG) << "StopTimeline is called. Blocking thread since "
                     "pending_status is still true.\n";
     } else {
@@ -66,7 +70,7 @@ std::string TimelineWriter::PendingTimelineFile() {
   return new_pending_filename_;
 }
 
-void TimelineWriter::SetTimelineFile(std::string filename) {
+void TimelineWriter::SetTimelineFile(const std::string& filename) {
   // No op if there are pending events in record_queue, let all the event from
   // record queue to get dumped to file
 
@@ -74,7 +78,7 @@ void TimelineWriter::SetTimelineFile(std::string filename) {
             << " New filename:" << filename;
 
   // Close if there existing file open and new file is not same as existing file
-  if (cur_filename_ != "" && cur_filename_ != filename) {
+  if (!cur_filename_.empty() && cur_filename_ != filename) {
 
     if (!record_queue_.empty()) {
       LOG(DEBUG) << " SetTimelineFile is no-op as there are events in "
@@ -92,7 +96,7 @@ void TimelineWriter::SetTimelineFile(std::string filename) {
   }
   // if new filename is empty, we need to stop accepting activities. This would
   // stopping timeline
-  if (filename == "") {
+  if (filename.empty()) {
 
     healthy_.exchange(1);
     active_.exchange(0);
@@ -126,7 +130,8 @@ void TimelineWriter::SetTimelineFile(std::string filename) {
 }
 
 void TimelineWriter::Initialize(
-    std::string file_name, std::chrono::steady_clock::time_point start_time_) {
+    const std::string& file_name,
+    std::chrono::steady_clock::time_point start_time_) {
   std::lock_guard<std::recursive_mutex> guard(writer_mutex_);
   if (healthy())
     return;
@@ -157,7 +162,7 @@ void TimelineWriter::Shutdown() {
               << e.code() << " meaning " << e.what();
   }
 
-  if (cur_filename_ != "" && file_.is_open()) {
+  if (!cur_filename_.empty() && file_.is_open()) {
     file_.flush();
     file_.close();
   }
@@ -262,7 +267,7 @@ void TimelineWriter::DoWriteEvent(const TimelineRecord& r) {
   if (r.phase == 'X') {
     file_ << ", \"dur\": " << 0 << "";
   }
-  if (r.args != "") {
+  if (!r.args.empty()) {
     file_ << ", \"args\": {" << r.args << "}";
   }
   // We make sure that the events are written always produce valid json file
@@ -324,7 +329,119 @@ void TimelineWriter::WriterLoop() {
   }
 }
 
-void Timeline::Initialize(std::string file_name, unsigned int horovod_size) {
+#if HAVE_NVTX
+class TimelineNvtxHandle {
+public:
+  TimelineNvtxHandle()
+      : main_domain_(nvtxDomainCreateA("HorovodTimeline")),
+        activity_domain_(nvtxDomainCreateA("HorovodTimelineActivities"))
+  { }
+
+  ~TimelineNvtxHandle() {
+    Disable();
+  }
+
+  void StartRange(const std::string& tensor_name,
+                  const std::string& event_category,
+                  int64_t tensor_size_payload = -1) {
+    if (main_domain_ == nullptr) {
+      return;
+    }
+    const auto message = event_category + " " + tensor_name;
+    nvtxRangeId_t range_id = NvtxStartDomainRange(main_domain_, message,
+                                                  tensor_size_payload);
+    main_tensor_nvtx_range_.emplace(tensor_name, range_id);
+  }
+
+  void EndRange(const std::string& tensor_name) {
+    if (main_domain_ == nullptr) {
+      return;
+    }
+    auto iter = main_tensor_nvtx_range_.find(tensor_name);
+    if (iter != main_tensor_nvtx_range_.end()) {
+      auto range_id = iter->second;
+      nvtxDomainRangeEnd(main_domain_, range_id);
+      main_tensor_nvtx_range_.erase(iter);
+    }
+  }
+
+  void StartActivityRange(const std::string& tensor_name,
+                          const std::string& activity) {
+    if (activity_domain_ == nullptr) {
+      return;
+    }
+    const auto message = activity + " " + tensor_name;
+    nvtxRangeId_t range_id = NvtxStartDomainRange(activity_domain_, message);
+    activity_tensor_nvtx_range_.emplace(tensor_name, range_id);
+  }
+
+  void EndActivityRange(const std::string& tensor_name) {
+    if (activity_domain_ == nullptr) {
+      return;
+    }
+    auto iter = activity_tensor_nvtx_range_.find(tensor_name);
+    if (iter != activity_tensor_nvtx_range_.end()) {
+      auto range_id = iter->second;
+      nvtxDomainRangeEnd(activity_domain_, range_id);
+      activity_tensor_nvtx_range_.erase(iter);
+    }
+  }
+
+  void Disable() {
+    if (main_domain_ != nullptr) {
+      nvtxDomainDestroy(main_domain_);
+      main_domain_ = nullptr;
+    }
+    if (activity_domain_ != nullptr) {
+      nvtxDomainDestroy(activity_domain_);
+      activity_domain_ = nullptr;
+    }
+  }
+
+private:
+  static nvtxRangeId_t NvtxStartDomainRange(nvtxDomainHandle_t domain,
+                                            const std::string& message,
+                                            int64_t tensor_size_payload = -1) {
+    nvtxEventAttributes_t eventAttrib = {0};
+    eventAttrib.version = NVTX_VERSION;
+    eventAttrib.size = NVTX_EVENT_ATTRIB_STRUCT_SIZE;
+    eventAttrib.messageType = NVTX_MESSAGE_TYPE_ASCII;
+    eventAttrib.message.ascii = message.c_str();
+    if (tensor_size_payload >= 0) {
+      eventAttrib.payloadType = NVTX_PAYLOAD_TYPE_INT64;
+      eventAttrib.payload.llValue = tensor_size_payload;
+    }
+
+    nvtxRangeId_t range_id = nvtxDomainRangeStartEx(domain, &eventAttrib);
+    // contents of eventAttrib and eventAtrrib.message.ascii have been copied
+    return range_id;
+  }
+
+  nvtxDomainHandle_t main_domain_;      // nullptr if disabled
+  nvtxDomainHandle_t activity_domain_;  // nullptr if disabled
+
+  std::unordered_map<std::string, nvtxRangeId_t> main_tensor_nvtx_range_;
+  std::unordered_map<std::string, nvtxRangeId_t> activity_tensor_nvtx_range_;
+};
+#else // HAVE_NVTX
+class TimelineNvtxHandle {
+public:
+  void StartRange(const std::string& tensor_name,
+                  const std::string& event_category,
+                  int64_t tensor_size_payload = -1) {}
+  void EndRange(const std::string& tensor_name) {}
+  void StartActivityRange(const std::string& tensor_name,
+                          const std::string& activity) {}
+  void EndActivityRange(const std::string& tensor_name) {}
+  void Disable() {}
+};
+#endif // HAVE_NVTX
+
+
+Timeline::Timeline() : nvtx_handle_(new TimelineNvtxHandle) {}
+
+void Timeline::Initialize(const std::string& file_name,
+                          unsigned int horovod_size) {
   if (Initialized()) {
     return;
   }
@@ -350,6 +467,8 @@ void Timeline::Shutdown() {
   tensor_states_.clear();
 }
 
+Timeline::~Timeline() = default;
+
 long Timeline::TimeSinceStartMicros() const {
   auto now = std::chrono::steady_clock::now();
   auto ts = now - start_time_;
@@ -393,6 +512,7 @@ void Timeline::NegotiateStart(const std::string& tensor_name,
 
   assert(tensor_states_[tensor_name] == TimelineState::UNKNOWN);
   auto event_category = "NEGOTIATE_" + Request::RequestType_Name(request_type);
+  nvtx_handle_->StartRange(tensor_name, event_category);
   WriteEvent(tensor_name, 'B', event_category);
   tensor_states_[tensor_name] = TimelineState::NEGOTIATING;
 }
@@ -415,12 +535,14 @@ void Timeline::NegotiateEnd(const std::string& tensor_name) {
 
   std::lock_guard<std::recursive_mutex> guard(mutex_);
   assert(tensor_states_[tensor_name] == TimelineState::NEGOTIATING);
+  nvtx_handle_->EndRange(tensor_name);
   WriteEvent(tensor_name, 'E');
   tensor_states_.erase(tensor_name);
 }
 
 void Timeline::Start(const std::string& tensor_name,
-                     const Response::ResponseType response_type) {
+                     const Response::ResponseType response_type,
+                     int64_t tensor_size) {
   if (!Initialized() || !writer_.active()) {
     return;
   }
@@ -429,6 +551,7 @@ void Timeline::Start(const std::string& tensor_name,
   assert(tensor_states_[tensor_name] == TimelineState::UNKNOWN);
   auto event_category = Response::ResponseType_Name(response_type);
   WriteEvent(tensor_name, 'B', event_category);
+  nvtx_handle_->StartRange(tensor_name, event_category, tensor_size);
   tensor_states_[tensor_name] = TimelineState::TOP_LEVEL;
 }
 
@@ -450,6 +573,7 @@ void Timeline::ActivityStart(const std::string& tensor_name,
 
   std::lock_guard<std::recursive_mutex> guard(mutex_);
   assert(tensor_states_[tensor_name] == TimelineState::TOP_LEVEL);
+  nvtx_handle_->StartActivityRange(tensor_name, activity);
   WriteEvent(tensor_name, 'B', activity);
   tensor_states_[tensor_name] = TimelineState::ACTIVITY;
 }
@@ -467,12 +591,13 @@ void Timeline::ActivityEnd(const std::string& tensor_name) {
 
   std::lock_guard<std::recursive_mutex> guard(mutex_);
   assert(tensor_states_[tensor_name] == TimelineState::ACTIVITY);
+  nvtx_handle_->EndActivityRange(tensor_name);
   WriteEvent(tensor_name, 'E');
   tensor_states_[tensor_name] = TimelineState::TOP_LEVEL;
 }
 
 void Timeline::End(const std::string& tensor_name,
-                   const std::shared_ptr<Tensor> tensor) {
+                   const std::shared_ptr<Tensor>& output_tensor) {
   if (!Initialized() || !writer_.active()) {
     return;
   }
@@ -484,12 +609,15 @@ void Timeline::End(const std::string& tensor_name,
     ActivityEnd(tensor_name);
   }
 
+  nvtx_handle_->EndRange(tensor_name);
+
   std::stringstream args;
-  if (tensor != nullptr) {
-    args << "\"dtype\": \"" << DataType_Name(tensor->dtype()) << "\"";
-    args << ", \"shape\": \"" << tensor->shape().DebugString() << "\"";
+  if (output_tensor != nullptr) {
+    args << "\"dtype\": \"" << DataType_Name(output_tensor->dtype()) << "\"";
+    args << ", \"shape\": \"" << output_tensor->shape().DebugString() << "\"";
   }
   WriteEvent(tensor_name, 'E', "", args.str());
+  tensor_states_.erase(tensor_name);
 }
 
 void Timeline::MarkCycleStart() {
@@ -501,9 +629,14 @@ void Timeline::MarkCycleStart() {
   WriteMarker("CYCLE_START");
 }
 
-void Timeline::SetPendingTimelineFile(std::string filename) {
+void Timeline::SetPendingTimelineFile(const std::string& filename) {
   writer_.SetPendingTimelineFile(filename);
   LOG(INFO) << "Set pending timeline file to " << filename;
 }
+
+void Timeline::DisableNvtx() {
+  nvtx_handle_->Disable();
+}
+
 } // namespace common
 } // namespace horovod
