@@ -517,6 +517,7 @@ def parse_args():
     config_parser.validate_config_args(args)
 
     args.run_func = None
+    args.executable = None
 
     if args.check_build:
         check_build(args.verbose)
@@ -598,7 +599,8 @@ def _run_static(args):
         put_data_into_kvstore(driver_ip, run_func_server_port,
                               'runfunc', 'func', args.run_func)
 
-        command = [sys.executable, '-m', 'horovod.runner.run_task', str(driver_ip), str(run_func_server_port)]
+        executable = args.executable or sys.executable
+        command = [executable, '-m', 'horovod.runner.run_task', str(driver_ip), str(run_func_server_port)]
 
         try:
             _launch_job(args, settings, nics, command)
