@@ -29,14 +29,14 @@ from horovod.tensorflow.functions import allgather_object, broadcast_object, bro
 from horovod.tensorflow.mpi_ops import allgather, broadcast, _allreduce, _grouped_allreduce, alltoall
 from horovod.tensorflow.mpi_ops import init, shutdown
 from horovod.tensorflow.mpi_ops import is_initialized, start_timeline, stop_timeline
-from horovod.tensorflow.mpi_ops import size, local_size, rank, local_rank, is_homogeneous
+from horovod.tensorflow.mpi_ops import size, local_size, cross_size, rank, local_rank, cross_rank, is_homogeneous
 from horovod.tensorflow.mpi_ops import rank_op, local_rank_op, size_op, local_size_op
 from horovod.tensorflow.mpi_ops import mpi_threads_supported, mpi_enabled, mpi_built
 from horovod.tensorflow.mpi_ops import gloo_enabled, gloo_built
 from horovod.tensorflow.mpi_ops import nccl_built, ddl_built, ccl_built, cuda_built, rocm_built
 from horovod.tensorflow.mpi_ops import Average, Sum, Adasum
 from horovod.tensorflow.mpi_ops import handle_average_backwards_compatibility, check_num_rank_power_of_2
-from horovod.tensorflow.util import _executing_eagerly, _make_subgraph, _cache
+from horovod.tensorflow.util import _executing_eagerly, _make_subgraph, _cache, vars_to_refs, refs_to_vars
 from horovod.tensorflow.mpi_ops import join
 from horovod.tensorflow.sync_batch_norm import SyncBatchNormalization
 from horovod.tensorflow.gradient_aggregation import LocalGradientAggregationHelper
@@ -331,9 +331,10 @@ if _SessionRunHook is not None and _get_default_graph is not None:
 
 
 @_cache
-def _make_allreduce_grads_fn(name, device_dense, device_sparse,
-                             compression, sparse_as_dense, op, gradient_predivide_factor,
-                             num_groups):
+def _make_cached_allreduce_grads_fn(name, device_dense, device_sparse,
+                                    compression, sparse_as_dense, op,
+                                    gradient_predivide_factor, groups):
+    groups = refs_to_vars(groups) if isinstance(groups, tuple) else groups
     if op == Average:
         # Split average operation across pre/postscale factors
         # C++ backend will apply additional 1 / size() factor to postscale_factor for op == Average.
@@ -343,26 +344,47 @@ def _make_allreduce_grads_fn(name, device_dense, device_sparse,
         prescale_factor = 1.0
         postscale_factor = 1.0
 
-    def allreduce_grads(grads):
+    def allreduce_grads(grads, vars=None):
         with tf.name_scope(name + "_Allreduce"):
             if sparse_as_dense:
                 grads = [tf.convert_to_tensor(grad)
                          if grad is not None and isinstance(grad, tf.IndexedSlices)
                          else grad for grad in grads]
 
-            if num_groups > 0:
-                grads_clean = [grad for grad in grads if grad is not None]
-                grads_split = split_list(grads_clean, num_groups)
+            if groups is not None:
+                if isinstance(groups, list):
+                    var_name2grad = {}
+                    for i in range(len(vars)):
+                        var = vars[i]
+                        grad = grads[i]
+                        if grad is not None:
+                            var_name2grad[var.name] = (i, grad)
+                    grads_split = []
+                    for group in groups:
+                        grad_group = []
+                        for var in group:
+                            if var.name in var_name2grad:
+                                grad_group.append(var_name2grad[var.name])
+                                del var_name2grad[var.name]
+                        grads_split.append(grad_group)
+                    for _, grad in var_name2grad.items():
+                        grads_split.append([grad])
+                elif groups > 0:
+                    grads_clean = [(i, grad) for i, grad in enumerate(grads) if grad is not None]
+                    grads_split = split_list(grads_clean, groups)
 
-                reduce_ops = []
+                reduce_ops = [None] * len(vars)
                 for group in grads_split:
-                     reduce_ops += _grouped_allreduce_cond(group,
-                                                          device_dense=device_dense,
-                                                          device_sparse=device_sparse,
-                                                          compression=compression,
-                                                          op=op,
-                                                          prescale_factor=prescale_factor,
-                                                          postscale_factor=postscale_factor)
+                    index_group, grad_group = [list(t) for t in zip(*group)]
+                    reduce_ops_group = _grouped_allreduce_cond(grad_group,
+                                                               device_dense=device_dense,
+                                                               device_sparse=device_sparse,
+                                                               compression=compression,
+                                                               op=op,
+                                                               prescale_factor=prescale_factor,
+                                                               postscale_factor=postscale_factor)
+                    for i in range(len(index_group)):
+                        reduce_ops[index_group[i]] = reduce_ops_group[i]
                 return reduce_ops
 
             return [_allreduce_cond(grad,
@@ -381,6 +403,15 @@ def _make_allreduce_grads_fn(name, device_dense, device_sparse,
         return allreduce_grads
 
 
+def _make_allreduce_grads_fn(name, device_dense, device_sparse,
+                             compression, sparse_as_dense, op,
+                             gradient_predivide_factor, groups):
+    groups = vars_to_refs(groups) if isinstance(groups, list) else groups
+    return _make_cached_allreduce_grads_fn(name, device_dense, device_sparse,
+                                           compression, sparse_as_dense, op,
+                                           gradient_predivide_factor, groups)
+
+
 try:
     # TensorFlow 2.x
     _LegacyOptimizer = tf.compat.v1.train.Optimizer
@@ -401,7 +432,7 @@ if _LegacyOptimizer is not None:
                     device_sparse='', compression=Compression.none,
                     sparse_as_dense=False, op=Average, gradient_predivide_factor=1.0,
                     backward_passes_per_step=1, average_aggregated_gradients=False,
-                    num_groups=0):
+                    groups=None):
             if name is None:
                 name = "Distributed{}".format(type(optimizer).__name__)
             super(_DistributedOptimizer, self).__init__(name=name, use_locking=use_locking)
@@ -409,7 +440,7 @@ if _LegacyOptimizer is not None:
             self._optimizer = optimizer
             self._allreduce_grads = _make_allreduce_grads_fn(
                 name, device_dense, device_sparse, compression, sparse_as_dense, op,
-                gradient_predivide_factor, num_groups)
+                gradient_predivide_factor, groups)
 
             self._agg_helper = None
             if backward_passes_per_step > 1:
@@ -439,22 +470,24 @@ if _LegacyOptimizer is not None:
             gradients = self._optimizer.compute_gradients(*args, **kwargs)
             grads, vars = zip(*gradients)
             if self._agg_helper:
-                avg_grads = self._agg_helper.compute_gradients(grads)
+                avg_grads = self._agg_helper.compute_gradients(grads, vars)
             else:
-                avg_grads = self._allreduce_grads(grads)
+                avg_grads = self._allreduce_grads(grads, vars)
             return list(zip(avg_grads, vars))
 
-        def apply_gradients(self, *args, **kwargs):
+        def apply_gradients(self, grads_and_vars, global_step=None, name=None):
             """Calls this same method on the underlying optimizer."""
             if self._agg_helper:
                 return self._agg_helper.apply_gradients(
-                    lambda: self._optimizer.apply_gradients(*args, **kwargs),
+                    lambda: self._optimizer.apply_gradients(
+                        grads_and_vars, global_step=global_step, name=name),
                     self._optimizer,
-                    *args,
-                    **kwargs,
+                    grads_and_vars,
+                    global_step=global_step,
+                    name=name,
                 )
-
-            return self._optimizer.apply_gradients(*args, **kwargs)
+            return self._optimizer.apply_gradients(
+                grads_and_vars, global_step=global_step, name=name)
 
         def get_slot(self, *args, **kwargs):
             """Calls this same method on the underlying optimizer."""
@@ -570,7 +603,7 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
                          sparse_as_dense=False, backward_passes_per_step=1,
                          op=Average, gradient_predivide_factor=1.0,
                          average_aggregated_gradients=False,
-                         num_groups=0):
+                         num_groups=0, groups=None):
     """Construct a new DistributedOptimizer, which uses another optimizer
     under the hood for computing single-process gradient values and
     applying gradient updates after the gradient values have been combined
@@ -619,6 +652,15 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
       num_groups:
         Number of groups to assign gradient allreduce ops to for explicit
         grouping. Defaults to no explicit groups.
+      groups:
+        The parameter to group the gradient allreduce ops. Accept values is a
+        non-negative integer or a list of list of tf.Variable.
+        If groups is a non-negative integer, it is the number of groups to assign
+        gradient allreduce ops to for explicit grouping.
+        If groups is a list of list of tf.Variable. Variables in the same
+        inner list will be assigned to the same group, while parameter that does
+        not appear in any list will form a group itself.
+        Defaults as None, which is no explicit groups.
     """
     if gradient_predivide_factor != 1.0:
         if rocm_built():
@@ -629,6 +671,17 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
     if op == Adasum and average_aggregated_gradients:
         raise ValueError('Adasum does not support average_aggregated_gradients == True')
 
+    if num_groups != 0:
+        warnings.warn('Parameter `num_groups` has been replaced by `groups` '
+                      'and will be removed in v0.23.0.', DeprecationWarning)
+        if groups is None:
+            groups = num_groups
+
+    if groups is not None:
+        if not (isinstance(groups, list) or groups > 0):
+            raise ValueError('groups should be a non-negative integer or '
+                            'a list of list of tf.Variable.')
+
     if isinstance(optimizer, _LegacyOptimizer):
         if op == Adasum:
             return _DistributedAdasumOptimizer(optimizer, name, use_locking, device_dense,
@@ -646,7 +699,7 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
             gradient_predivide_factor=gradient_predivide_factor,
             backward_passes_per_step=backward_passes_per_step,
             average_aggregated_gradients=average_aggregated_gradients,
-            num_groups=num_groups
+            groups=groups
         )
     elif isinstance(optimizer, tf.keras.optimizers.Optimizer):
         if op == Adasum:
@@ -672,7 +725,8 @@ def DistributedOptimizer(optimizer, name=None, use_locking=False, device_dense='
 if hasattr(tf, 'GradientTape'):
     class _DistributedGradientTape(tf.GradientTape):
         def __init__(self, tape, device_dense, device_sparse, compression, sparse_as_dense, op,
-                     gradient_predivide_factor, num_groups, persistent=False, watch_accessed_variables=True):
+                     gradient_predivide_factor, groups, persistent=False,
+                     watch_accessed_variables=True):
             if hasattr(tape, '_watch_accessed_variables'):
                 super(self.__class__, self).__init__(persistent, watch_accessed_variables)
             else:
@@ -681,17 +735,17 @@ if hasattr(tf, 'GradientTape'):
             self._tape = tape
             self._allreduce_grads = _make_allreduce_grads_fn(
                 'DistributedGradientTape', device_dense, device_sparse, compression,
-                sparse_as_dense, op, gradient_predivide_factor, num_groups)
+                sparse_as_dense, op, gradient_predivide_factor, groups)
 
         def gradient(self, target, sources, output_gradients=None):
             gradients = super(self.__class__, self).gradient(target, sources, output_gradients)
-            return self._allreduce_grads(gradients)
+            return self._allreduce_grads(gradients, sources)
 
 
     def DistributedGradientTape(gradtape, device_dense='', device_sparse='',
                                 compression=Compression.none, sparse_as_dense=False,
                                 op=Average, gradient_predivide_factor=1.0,
-                                num_groups=0):
+                                num_groups=0, groups=None):
         """A tape that wraps another tf.GradientTape, using an allreduce to
         combine gradient values before applying gradients to model weights.
 
@@ -723,6 +777,15 @@ if hasattr(tf, 'GradientTape'):
           num_groups:
             Number of groups to assign gradient allreduce ops to for explicit
             grouping. Defaults to no explicit groups.
+          groups:
+            The parameter to group the gradient allreduce ops. Accept values is a
+            non-negative integer or a list of list of tf.Variable.
+            If groups is a non-negative integer, it is the number of groups to assign
+            gradient allreduce ops to for explicit grouping.
+            If groups is a list of list of tf.Variable. Variables in the same
+            inner list will be assigned to the same group, while parameter that does
+            not appear in any list will form a group itself.
+            Defaults as None, which is no explicit groups.
         """
         if gradient_predivide_factor != 1.0:
             if rocm_built():
@@ -730,13 +793,24 @@ if hasattr(tf, 'GradientTape'):
             if op != Average:
                 raise ValueError('gradient_predivide_factor not supported with op != Average')
 
+        if num_groups != 0:
+            warnings.warn('Parameter `num_groups` has been replaced by `groups` '
+                          'and will be removed in v0.23.0.', DeprecationWarning)
+            if groups is None:
+                groups = num_groups
+
+        if groups is not None:
+            if not (isinstance(groups, list) or groups > 0):
+                raise ValueError('groups should be a non-negative integer or '
+                                'a list of list of tf.Variable.')
+
         cls = type(gradtape.__class__.__name__, (gradtape.__class__,),
                    dict(_DistributedGradientTape.__dict__))
         if hasattr(gradtape, '_watch_accessed_variables'):
             return cls(gradtape._tape, device_dense, device_sparse, compression,
-                       sparse_as_dense, op, gradient_predivide_factor, num_groups,
+                       sparse_as_dense, op, gradient_predivide_factor, groups,
                        gradtape._persistent, gradtape._watch_accessed_variables)
         else:
             return cls(gradtape._tape, device_dense, device_sparse, compression,
-                       sparse_as_dense, op, gradient_predivide_factor, num_groups,
+                       sparse_as_dense, op, gradient_predivide_factor, groups,
                        gradtape._persistent)
