@@ -30,7 +30,9 @@ from horovod.tensorflow import local_rank
 from horovod.tensorflow import mpi_threads_supported, mpi_enabled, mpi_built
 from horovod.tensorflow import gloo_enabled, gloo_built
 from horovod.tensorflow import nccl_built, ddl_built, ccl_built, cuda_built, rocm_built
-from horovod.tensorflow import Average, Compression, Sum
+from horovod.tensorflow import Average, Sum
+from horovod.tensorflow import compression
+
 
 import horovod._keras as _impl
 from horovod.tensorflow.keras import callbacks, elastic
@@ -47,12 +49,14 @@ except:
 
 def DistributedOptimizer(optimizer, name=None,
                          device_dense='', device_sparse='',
-                         compression=Compression.none,
+                         compression=compression.Compression.none,
                          sparse_as_dense=False,
                          gradient_predivide_factor=1.0,
                          op=Average,
                          backward_passes_per_step=1,
-                         average_aggregated_gradients=False):
+                         average_aggregated_gradients=False,
+                         num_groups=0,
+                         groups=None):
     """
     An optimizer that wraps another keras.optimizers.Optimizer, using an allreduce to
     average gradient values before applying gradients to model weights.
@@ -87,6 +91,16 @@ def DistributedOptimizer(optimizer, name=None,
                                       If true divides gradient updates by
                                       backward_passes_per_step.
                                       Only applicable for backward_passes_per_step > 1.
+        num_groups: Number of groups to assign gradient allreduce ops to for explicit
+                    grouping. Defaults to no explicit groups.
+        groups: The parameter to group the gradient allreduce ops. Accept values is a
+                non-negative integer or a list of list of tf.Variable.
+                If groups is a non-negative integer, it is the number of groups to assign
+                gradient allreduce ops to for explicit grouping.
+                If groups is a list of list of tf.Variable. Variables in the same
+                inner list will be assigned to the same group, while parameter that does
+                not appear in any list will form a group itself.
+                Defaults as None, which is no explicit groups.
     """
     if gradient_predivide_factor != 1.0 and rocm_built():
             raise ValueError('gradient_predivide_factor not supported yet with ROCm')
@@ -94,6 +108,17 @@ def DistributedOptimizer(optimizer, name=None,
     if op != Average and op != Sum:
         raise ValueError('op currently only supports Average and Sum')
 
+    if num_groups != 0:
+        warnings.warn('Parameter `num_groups` has been replaced by `groups` '
+                      'and will be removed in v0.23.0.', DeprecationWarning)
+        if groups is None:
+            groups = num_groups
+
+    if groups is not None:
+        if not (isinstance(groups, list) or groups > 0):
+            raise ValueError('groups should be a non-negative integer or '
+                            'a list of list of tf.Variable.')
+
     return _impl.create_distributed_optimizer(
         keras=keras,
         optimizer=optimizer,
@@ -106,6 +131,7 @@ def DistributedOptimizer(optimizer, name=None,
         op=op,
         backward_passes_per_step=backward_passes_per_step,
         average_aggregated_gradients=average_aggregated_gradients,
+        groups=groups,
     )
 
 
@@ -123,7 +149,7 @@ def allreduce(value, name=None, average=None,
               prescale_factor=1.0,
               postscale_factor=1.0,
               op=None,
-              compression=Compression.none):
+              compression=compression.Compression.none):
     """
     Perform an allreduce on a tensor-compatible value.
 
@@ -184,7 +210,7 @@ def broadcast(value, root_rank, name=None):
     return _impl.broadcast(K, value, root_rank, name)
 
 
-def load_model(filepath, custom_optimizers=None, custom_objects=None, compression=Compression.none):
+def load_model(filepath, custom_optimizers=None, custom_objects=None, compression=compression.Compression.none):
     """
     Loads a saved Keras model with a Horovod DistributedOptimizer.
 
