@@ -37,7 +37,7 @@ using GpuStreamHandle = cudaStream_t;
 #include <hip/hip_runtime.h>
 using GpuStreamHandle = hipStream_t;
 #define gpuMemsetAsync hipMemsetAsync
-#endif
+#endif // HAVE_CUDA, HAVE_ROCM
 
 // Forward declaration of AsGpuStreamValue
 namespace stream_executor {
@@ -46,7 +46,7 @@ GpuStreamHandle AsGpuStreamValue(Stream* stream);
 } // namespace stream_executor
 } // namespace gpu
 #include "tensorflow/stream_executor/stream.h"
-#endif
+#endif // HAVE_GPU
 
 #define OMPI_SKIP_MPICXX
 #include "../common/operations.h"
@@ -163,6 +163,9 @@ public:
   AllocateOutput(common::TensorShape shape,
                  std::shared_ptr<common::Tensor>* tensor) override;
   virtual common::Status
+  AllocateOutput(int output_index, common::TensorShape shape,
+                 std::shared_ptr<common::Tensor>* tensor) override;
+  virtual common::Status
   AllocateZeros(int64_t num_elements, common::DataType dtype,
                 std::shared_ptr<common::Tensor>* tensor) override;
   virtual common::Framework framework() const override;
@@ -276,12 +279,18 @@ common::Status TFOpContext::AllocatePersistent(
 common::Status
 TFOpContext::AllocateOutput(common::TensorShape shape,
                             std::shared_ptr<common::Tensor>* tensor) {
+  return TFOpContext::AllocateOutput(0, shape, tensor);
+}
+
+common::Status
+TFOpContext::AllocateOutput(int output_index, common::TensorShape shape,
+                            std::shared_ptr<common::Tensor>* tensor) {
   TensorShape tf_shape;
   for (int idx = 0; idx < shape.dims(); ++idx) {
     tf_shape.AddDim(shape.dim_size(idx));
   }
   Tensor* tf_tensor;
-  Status status = context_->allocate_output(0, tf_shape, &tf_tensor);
+  Status status = context_->allocate_output(output_index, tf_shape, &tf_tensor);
   if (status.ok()) {
     *tensor = std::make_shared<TFTensor>(*tf_tensor);
   }
@@ -495,26 +504,29 @@ public:
     int num_tensors = num_tensors_;
 
     for (int i = 0; i < num_tensors_; ++i) {
-        auto tensor = context->input(i);
-        OP_REQUIRES_OK_ASYNC(
-            context, context->allocate_output(i, tensor.shape(), &outputs[i]), done);
-        // ReadyEvent makes sure input tensor is ready, and output is allocated.
-        ready_events.emplace_back(std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
-        hvd_contexts.emplace_back(std::make_shared<TFOpContext>(context));
-        hvd_tensors.emplace_back(std::make_shared<TFTensor>(tensor));
-        names.emplace_back(node_name + "_" + std::to_string(i+1) + "of" + std::to_string(num_tensors));
-        hvd_outputs.emplace_back(std::make_shared<TFTensor>(*outputs[i]));
-        callbacks.emplace_back(
-            [context, done, callback_mutex, callback_count, num_tensors](const common::Status& status) mutable {
-                // Must only invoke callback on last tensor.
-                std::lock_guard<std::mutex> guard(*callback_mutex);
-                (*callback_count)++;
-                if (*callback_count == num_tensors) {
-                    context->SetStatus(ConvertStatus(status));
-                    done();
-                }
+      auto tensor = context->input(i);
+      OP_REQUIRES_OK_ASYNC(
+          context, context->allocate_output(i, tensor.shape(), &outputs[i]),
+          done);
+      // ReadyEvent makes sure input tensor is ready, and output is allocated.
+      ready_events.emplace_back(
+          std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+      hvd_contexts.emplace_back(std::make_shared<TFOpContext>(context));
+      hvd_tensors.emplace_back(std::make_shared<TFTensor>(tensor));
+      names.emplace_back(node_name + "_" + std::to_string(i + 1) + "of" +
+                         std::to_string(num_tensors));
+      hvd_outputs.emplace_back(std::make_shared<TFTensor>(*outputs[i]));
+      callbacks.emplace_back(
+          [context, done, callback_mutex, callback_count, num_tensors]
+          (const common::Status& status) {
+            // Must only invoke callback on last tensor.
+            std::lock_guard<std::mutex> guard(*callback_mutex);
+            (*callback_count)++;
+            if (*callback_count == num_tensors) {
+              context->SetStatus(ConvertStatus(status));
+              done();
             }
-        );
+          });
     }
 
     auto enqueue_result = EnqueueTensorAllreduces(
@@ -549,7 +561,6 @@ REGISTER_OP("HorovodGroupedAllreduce")
     .Input("tensors: num_tensors*T")
     .Output("sum: num_tensors*T")
     .SetShapeFn([](shape_inference::InferenceContext* c) {
-      int num_tensors;
       for (int i = 0; i < c->num_inputs(); ++i) {
           c->set_output(i, c->input(i));
       }
@@ -910,7 +921,10 @@ private:
 REGISTER_KERNEL_BUILDER(Name("HorovodAlltoall").Device(DEVICE_CPU),
                         HorovodAlltoallOp);
 #if HOROVOD_GPU_ALLTOALL
-REGISTER_KERNEL_BUILDER(Name("HorovodAlltoall").Device(DEVICE_GPU).HostMemory("splits"),
+REGISTER_KERNEL_BUILDER(Name("HorovodAlltoall")
+                            .Device(DEVICE_GPU)
+                            .HostMemory("splits")
+                            .HostMemory("received_splits"),
                         HorovodAlltoallOp);
 #endif
 
@@ -921,11 +935,13 @@ REGISTER_OP("HorovodAlltoall")
     .Input("tensor: T")
     .Input("splits: int32")
     .Output("output: T")
+    .Output("received_splits: int32")
     .SetShapeFn([](shape_inference::InferenceContext* c) {
       shape_inference::ShapeHandle output;
       TF_RETURN_IF_ERROR(
           c->ReplaceDim(c->input(0), 0, c->UnknownDim(), &output));
       c->set_output(0, output);
+      c->set_output(1, c->input(1));
       return Status::OK();
     })
     .Doc(R"doc(
@@ -933,11 +949,13 @@ Perform an MPI Alltoall on a tensor.
 
 Arguments
     tensor:     A tensor to be distributed with all to all
-    splits: A list of integers in rank order describing how many elements
+    splits:     A list of integers in rank order describing how many elements
                 in `tensor` to send to each worker.
 
 Output
-    output:    The collected tensor data from all workers.
+    output:           The collected tensor data from all workers.
+    received_splits:  A list of integers in rank order describing how many
+                      elements in `output` have been received from each worker.
 )doc");
 
 } // namespace tensorflow
